/*
Copyright (c) 2013-2014 Gustav Hartvigsson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#ifndef __H_DYNAMIC_ARRAY__
#define __H_DYNAMIC_ARRAY__
#include <stdlib.h>
#include "defs.h"
#include "Func.h"

BEGIN_DECLS

/** @file
 * @defgroup SDynamicArray SDynamicArray
 * @addtogroup SDynamicArray SDynamicArray
 * @{
 *
 * SDynamicArray is an imlpementation of a dynamic array, it is usefule when
 * dealing with lare amounts of data that may change over time, or with an
 * unknowned numebr of items.
 * 
 * Note that accsess time is constant, but write time is not guarenteed to be.
 * 
 * When the size of the array is equal to the number of elements in it, it will
 * re-allocate the array with a larger size.
 */

/**
 * The padding that is added when expanding the array.
 */
#define ARRAY_PADDING 8

/**
 * An SDynamicArray is the standard implementation of a Dynamic Array in SSTS.
 *
 * It does not depend an SObject, because it should be albe to be used in,
 * SObject and SObject based classes.
 */
typedef struct SDynamicArray SDynamicArray;


#define S_DYNAMIC_ARRAY(k) ((SDynamicArray *)(k))

/**
 * Create a new dynamic array.
 *
 * @param len The length of the initial array.
 * @param free_func The function to be used when freeing the items. Can be NULL.
 *        If free_func is NULL, it will use the standard library's 
 *       <tt>free()</tt>.
 *
 * The free function should have the the signature <tt> (DynamicArray * self,
 * spointer item, spointer data) </tt> should be used but cast to FreeFunc.
 */
SDynamicArray *
s_dynamic_array_new (size_t len,
                     FreeFunc free_func);

/**
 * Same as s_dynamic_array_new() but with support for to_json and from_json
 * methods.
 *
 * @param to_json This function is used, on an item basis, to serialise the data
 *        in that item into a json representation.
 *
 * @param from_json This function talkes a string and marchal it into an object.
 *
 * The to_json function should have the signature:
 * <tt>char * name (spointer item)</tt> and return a string.
 * If this is not set, it will return <tt>"(pointer)"</tt>.
 *
 * The from_json function should have the signature:
 * <tt>spointer name (char * json)</tt>.
 * If this function is not set, the s_matrix_deserialize_json() function will
 * not work.
 */
SDynamicArray *
s_dynamic_array_new_json (size_t len,
                          FreeFunc free_func,
                          FuncPointer to_json,
                          FuncPointer from_json);

/**
 * Frees the dynamic array.
 * 
 * after this is run the data will be lost.
 */
void
s_dynamic_array_free (SDynamicArray * self, sboolean free_data);


/**
 * Get an item from the array.
 */
spointer
s_dynamic_array_get (SDynamicArray * self, size_t index);

/**
 *
 */
void
s_dynamic_array_set (SDynamicArray * self, size_t index, spointer item);

/**
 * Get the size of the array, this is not the same as the length of the array.
 * The size is the number of elements that can be allocated without resizing
 * the array.
 * 
 * To get the length of the array use s_dynamic_array_len ().
 */
size_t
s_dynamic_array_size (SDynamicArray * self);

/**
 * Get the index of the last item in the array.
 *
 * @note
 * This is not the last item added to the array, necessary, it is the index
 * of the item that has the highest position in the array.
 */
size_t
s_dynamic_array_last_item (SDynamicArray * self);

/**
 * Dumps a copy of the array. Must be cast.
 * 
 * Is not freed when the dynamic array is freed.
 * 
 * Is null-terminated.
 */
spointer *
s_dynamic_array_dump_array (SDynamicArray * self);

/**
 * Use a function on the array.
 */
void
s_dynamic_array_for_each (SDynamicArray * self, ForEachFunc func,
                          spointer data);

#if 0
This function is removed, it would have been a pain in the arse to implement.
/** TODO
 * same as s_dynamic_array_for_each (), with the difference that it returns a new
 * SDynamicArray.
 */
SDynamicArray *
s_dynamic_array_for_each_with_return (SDynamicArray * self,
                                      ForEachFunc func,
                                      spointer data);
#endif

/**
 * Returns a JSON representation of the array.
 *
 * If the to_json function is not set, this will return <tt>(pointer)</tt> for
 * each object.
 *
 * @param self The SDynamicArray to get the JSON representation from.
 *
 * @return a string with the JSON representation of the SDynamicArray.
 */
char *
s_dynamic_array_serialize_json (SDynamicArray * self);

/**
 * appends itmes to an the array 
 *
 * If the from_json function is not set, this will couse undefined behaviour.
 */
void
s_dynamic_array_deserialize_json (SDynamicArray * self, char * json);

/** @} */

END_DECLS

#endif /* #define __H_DYNAMIC_ARRAY__ */
