/*
Copyright (c) 2013-2014 Gustav Hartvigsson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

/** @file */

#ifndef __H_ERROR__
#define __H_ERROR__

#include "baseobject.h"
#include "defs.h"
#include "utils.h"
#include "Func.h"
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

BEGIN_DECLS

/**
 * @defgroup SError SError
 * @addtogroup SError
 * @{
 * An opaque datastructure that holds the SError's private data.
 */
typedef struct SErrorPrivate SErrorPrivate;


typedef void (* SErrorDomainToString)(sint error_id, schar * error_msg);

/**
 * An SError is a data structure that inherits from SBaseObjectInstance.
 * 
 * An SError represents an error that can occur.
 */
typedef struct
SError {
  SObject parent;
  SErrorPrivate * priv; /** Pimple pointer to the private data. */
  
} SError;

/**
 * The error class is not changed, no extra methods are needed.
 */
typedef struct
SErrorClass {
  SObjectClass parent_class;
  
} SErrorClass;

#define S_ERROR(o) (SError *)(o)
#define S_ERROR_CLASS(k) (SErrorClass *)(k)

/** @brief
 * The different error types.
 */
typedef enum {
  S_ERROR_NONE = 0, /**< Not on error */
  S_ERROR_INPUT_OUTPUT,/**< An I/O error */
  S_ERROR_OVERFLOW, /**< An Overflow error */
  S_ERROR_TYPE_ERROR, /** Some generic type error. */
  S_ERROR_UNUSED_0,
  S_ERROR_UNUSED_1,
  S_ERROR_UNUSED_2,
  S_ERROR_OTHER, /**< Some unknowned error */
  S_ERROR_NULL /**< An NULL error */
} SErrorType;

/**
 * @var static char * SErrorTypeName[]
 * @breif
 * Holds the names of the string version of SErrorType.
 */
static schar * SErrorTypeName[] UNUSED = {
  "NONE",
  "INPUT/OUTPUT",
  "OVERFLOW",
  "TYPE_ERROR",
  "UNUSED_0/INVALID",
  "UNUSED_1/INVALID",
  "UNUSED_2/INVALID",
  "OTHER",
  "NULL",
  0x0,
  0x0
};



/**
 * @brief
 * Function to get the same of an error.
 * 
 * For use in bindings.
 *
 * @param k The key to look up.
 */
schar *
s_error_get_name (SErrorType k);

/**
 * the constructor for the an SError
 */
SError *
s_error_new (SErrorType error, schar * message);

/**
 * This function calles the deinitize method of the object.
 */
void
s_error_free (SError * self);

/**
 * This function returns the ErrorType of on object.
 */
SErrorType
s_error_get_error_type (SError * self);

/**
 * This function prints the current error to stdout.
 */
void
s_error_print_error (SError * self);

#if 0
TODO
##########

I have been thinking about Error Propagation in SSTS and GLib.

I noticed that errors in GLib does not rely propagate, you can do 
g_error_propagate () but that assumes that the error that is fed to the function
is an array. This assumption can never be made and makes it very complex to deal
with different ways of doing things in different parts of a program (GError * vs
GError **).

So what I am thinking of doing is a little more complex: have explicit 
registration of error domains (s_error_domain_register (schar * name,
FuncPointer to_string), s_error_domain_is_registered (schar *)).

The other thing I was thinking is that errors should have implicit propagation.
(IE: no s_error_new () only s_error_append ()) and have the propagation
information stored in an array. (ID and message).

The problem with this is that where do you call s_error_domain_register ()?

In the constructor of an SObject would be a great place to put it, if not for
the problem that we have to call s_error_domain_is_registered when a new
instance of an object is created.

In the main function is just tedious, Wrapping the main function in something
like an S_MAIN macro could work and have an internal
s_library_register_error_domains function somewhere and require the user to
implement an s_user_register_error_domains function.

This puts even more work on the shoulders of the developers that uses the
library.

It seems that no matter what I do (A true propagating error system or
something like GError from GLib) I put extra work in the lap of the developer
that uses the library.﻿

Commets?

##########

void
s_error_domain_register (schar * domain, SErrorDomainToString func);

sboolean
s_error_domain_is_registerd (schar * domain);
#endif

/** @} */

END_DECLS

#endif
