/*
Copyright (c) 2013-2014 Gustav Hartvigsson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include <assert.h>
#include "LinkedList.h"
#include "stdio.h"

struct
SLinkedListNode {
  SLinkedListNode * prev;
  SLinkedListNode * next;
  spointer data;
};

struct
SLinkedList {
  size_t len; /**< The length of the list, may be wrong.*/
  FreeFunc free_func; /**< The function that frees the data. */
  SLinkedListNode * head; /**< head->prev must be NULL */
  SLinkedListNode * tail; /**< tail->next must be NULL */
  SLinkedListNode * current; /**< Represents the curret object*/
};



SLinkedList *
s_linked_list_new (FreeFunc free_func) {
  SLinkedList * self = malloc (sizeof (SLinkedList));
  self->len = 0;
  self->free_func = free_func;
  self->head = self->tail = self->current = NULL;
  return self;
}

void
s_linked_list_free (SLinkedList * self, sboolean free_data) {
  
}

void
internal_s_linked_list_initialize_first_item (SLinkedList * self) {
  assert (self->len == 0);
  assert (self->head == NULL && self->tail == NULL && self->current == NULL);
  self->head = self->tail = self->current = malloc (sizeof (SLinkedListNode));
  self->head->prev = NULL;
  self->tail->next = NULL;
}

void
s_linked_list_add (SLinkedList * self, spointer data) {
  if (self->len  == 0) { // special case when the list is empty.
    internal_s_linked_list_initialize_first_item (self);
  } else { // general case.
    SLinkedListNode * old_current = self->current;
    SLinkedListNode * next = old_current->next;
    SLinkedListNode * new_current = malloc (sizeof (SLinkedListNode));
    
    new_current->next = next;
    new_current->prev = old_current;
    old_current->next = new_current;
    next->prev = new_current;
    
    self->current = new_current;
  }
  self->current->data = data;
  self->len = self->len + 1;
}

void
s_linked_list_append (SLinkedList * self, spointer data) {
  if (self->len  == 0) { // special case when the list is empty.
    internal_s_linked_list_initialize_first_item (self);
  } else { // general case.
    self->tail->next = malloc (sizeof (SLinkedListNode));
    self->tail->next->next = NULL;
    self->tail = self->tail->next;
  }
  self->tail->data = data;
  self->len = self->len + 1;
  
  assert (self->tail->next == NULL);
}

void
s_linked_list_prepend (SLinkedList * self, spointer data) {
  if (self->len  == 0) { // special case when the list is empty.
    internal_s_linked_list_initialize_first_item (self);
  } else { // general case.
    self->head->prev = malloc (sizeof (SLinkedListNode));
    self->head->prev->prev = NULL;
    self->head = self->head->prev;
  }
  self->head->data = data;
  self->len = self->len + 1;
  
  assert (self->head->prev == NULL);
}

sboolean
s_linked_list_next (SLinkedList * self) {
  if (self->current->next) {
    self->current = self->current->next;
    return TRUE;
  } else {
    assert (self->current == self->tail);
    return FALSE;
  }
}


sboolean
s_linked_list_prev (SLinkedList * self) {
  if (self->current->prev) {
    self->current = self->current->prev;
    return TRUE;
  } else {
    assert (self->current == self->head);
    return FALSE;
  }
}


spointer
s_linked_list_get_current (SLinkedList * self) {
  return self->current->data;
}


spointer
s_linked_list_get_next (SLinkedList * self) {
  if (self->current->next) {
    return self->current->next->data;
  } else {
    assert (self->current == self->tail);
    return NULL;
  }
}


spointer
s_linked_list_get_prev (SLinkedList * self) {
    if (self->current->prev) {
    return self->current->next->data;
  } else {
    assert (self->current == self->head);
    return NULL;
  }
}


void
s_linked_list_head (SLinkedList * self) {
  self->current = self->head;
}


void
s_linked_list_tail (SLinkedList * self) {
  self->current = self->tail;
}


size_t
s_linked_list_len (SLinkedList * self) {
  return self->len;
}

void
s_linked_list_remove_current (SLinkedList * self, sboolean free_data) {
  SLinkedListNode * prev = self->current->prev;
  SLinkedListNode * next = self->current->next;
  SLinkedListNode * current = self->current;
  
  if (free_data) {
    FreeFunc free_func = self->free_func;
    spointer data = current->data;
    if (free_func != NULL) {
      free_func (data);
    } else {
      free (data);
    }
  }
  
  self->current = prev;
  
  prev->next = next;
  next->prev = prev;
  
  free (current);
  
  self->len = self->len - 1;
}

void
s_linked_list_for_each (SLinkedList * self, ForEachFunc func,
                        spointer user_data) {
  SLinkedListNode * current_pos = self->current;
  
  s_linked_list_head (self);
  
  
  if(self->current != NULL) {
    do {
      spointer cur_itm = s_linked_list_get_current (self);
      if (cur_itm) {
        func (self, cur_itm, user_data);
      }
    } while (s_linked_list_next (self));
  }
  self->current = current_pos;
}

